/* Render.CPP
 * ----------
 * Using the idea from MAME32, use several different render routines to solve 
 * the different screen depth problems, also getting rid of the dreaded tripple 
 * buffer idea when possible
 */

#include <memory.h>
#include "Render.h"
#include "snes9x\port.h"
#include "snes9x.h"
#include "snes9x\snes9x.h"

// Private Prototypes, should not be called directly
void RenderNothing( SSurface Src, SSurface Dst);
void Render8to8( SSurface Src, SSurface Dst);
void RenderDouble8to8( SSurface Src, SSurface Dst);
void RenderFakeTV8to8( SSurface Src, SSurface Dst);
void Render8to16( SSurface Src, SSurface Dst);
void RenderDouble8to16( SSurface Src, SSurface Dst);
void RenderFakeTV8to16( SSurface Src, SSurface Dst);
void Render16to16( SSurface Src, SSurface Dst);
void RenderDouble16to16( SSurface Src, SSurface Dst);
void RenderFakeTV16to16( SSurface Src, SSurface Dst);
void RenderSuperEagle16to16( SSurface Src, SSurface Dst);
void Render2xSaI16to16( SSurface Src, SSurface Dst);
void RenderSuper2xSaI16to16( SSurface Src, SSurface Dst);

// Contains the pointer to the now active render method
TRenderMethod OldRenderMethod;
TRenderMethod RenderMethod;


// Currently only 16-bit is supported, but this function will need to be 
// extended to also support 8-bit routines and check which routines should
// be used...
void SelectRenderMethod()
{
	OldRenderMethod = RenderMethod;
	RenderMethod = RenderNothing;

	if( (GUI.Depth == 8) && (Settings.SixteenBit == false))
	switch( GUI.Scale)
	{
		case 0:
			RenderMethod = Render8to8;
			break;
		case 1:
			RenderMethod = RenderDouble8to8;
			break;
		case 2:
			RenderMethod = RenderFakeTV8to8;
			break;
		default:
			RenderMethod = Render8to8;
			break;
	}

	if( (GUI.Depth == 16) && (Settings.SixteenBit == true))
	switch( GUI.Scale)
	{
		case 0: 
			RenderMethod = Render16to16;
			break;
		case 1: 
			RenderMethod = RenderDouble16to16;
			break;
		case 2: 
			RenderMethod = RenderFakeTV16to16;
			break;
		case 3: 
			RenderMethod = RenderSuperEagle16to16;
			break;
		case 4:
			RenderMethod = Render2xSaI16to16;
			break;
		case 5:
			RenderMethod = RenderSuper2xSaI16to16;
			break;
		default:
			RenderMethod = Render16to16;
			break;
	}

	if( (GUI.Depth == 16) && (Settings.SixteenBit == false))
	switch( GUI.Scale)
	{
		case 0: 
			RenderMethod = Render8to16;
			break;
		case 1: 
			RenderMethod = RenderDouble8to16;
			break;
		case 2: 
			RenderMethod = RenderFakeTV8to16;
			break;
		default:
			RenderMethod = Render8to16;
			break;
	}
}

void RenderNothing( SSurface Src, SSurface Dst)
{ // Just an empty procedure, so that it never happens that something 'bad' is rendered =)
}

// Needed by RenderDouble8to8, stretches a single line
inline void DoubleLine8( unsigned __int8 *lpDst, unsigned __int8 *lpSrc, unsigned int Width)
{
	for( unsigned int W = 0; W != Width; W ++)
	{
		lpDst[0] = lpSrc[0];
		lpDst[1] = lpSrc[0];

		lpDst += 2;
		lpSrc ++;
	}
}

void Render8to8( SSurface Src, SSurface Dst)
{
    unsigned char *lpSrc, *lpDst;
    bool ExtendHeight, DoubleHeight, DoubleWidth;

    lpSrc = Src.Surface;

    ExtendHeight = Src.Height == 224 || Src.Height == 448;
    DoubleHeight = Src.Height > 240;
    DoubleWidth  = Src.Width == 512;

    lpDst = Dst.Surface;
    lpDst += ((Dst.Height - (DoubleHeight ? 478 : 239)) >> 1) * Dst.Pitch;
    lpDst += (Dst.Width - Src.Width) >> 1;
	
    for( unsigned int H = 0; H != Src.Height; H ++, lpDst += Dst.Pitch, lpSrc += Src.Pitch)
    {	memcpy( lpDst, lpSrc, Src.Width);	}

    if( ExtendHeight)
    {
        unsigned int HH = DoubleHeight ? 478 : 239;
	for(; H < HH; H++, lpDst += Dst.Pitch)
	{	memset( lpDst, 0, Src.Width);	}
    }
}

void RenderDouble8to8( SSurface Src, SSurface Dst)
{
unsigned char *lpSrc, *lpDst;
bool ExtendHeight, DoubleHeight, DoubleWidth;
int H;

	lpSrc = Src.Surface;

	ExtendHeight = Src.Height == 224 || Src.Height == 448;
	DoubleHeight = Src.Height > 240;
	DoubleWidth  = Src.Width == 512;

	// If Snes9x is rendering anything in HiRes, then just copy, don't stretch
	if( DoubleHeight || DoubleWidth)
	{	
		Render16to16( Src, Dst);
		return;
	}

	lpDst = Dst.Surface;
	lpDst += ((Dst.Height - 478) >> 1) * Dst.Pitch;
    lpDst += (Dst.Width - 512) >> 1;

	for( H = 0; H != (ExtendHeight ? 448 : 478); H += 2)
	{
		DoubleLine8( (unsigned __int8 *)lpDst, (unsigned __int8 *)lpSrc, Src.Width);
		lpDst += Dst.Pitch;

		DoubleLine8( (unsigned __int8 *)lpDst, (unsigned __int8 *)lpSrc, Src.Width);
		lpSrc += Src.Pitch;
		lpDst += Dst.Pitch;
	}

	if( ExtendHeight)
	for(; H != 478; H += 1)
	{	
		memset( lpDst, 0, 512);
		lpDst += Dst.Pitch;
	}
}

void RenderFakeTV8to8( SSurface Src, SSurface Dst)
{
unsigned char *lpSrc, *lpDst;
bool ExtendHeight, DoubleHeight, DoubleWidth;
int H;

	lpSrc = Src.Surface;

	ExtendHeight = Src.Height == 224 || Src.Height == 448;
	DoubleHeight = Src.Height > 240;
	DoubleWidth  = Src.Width == 512;

	// If Snes9x is rendering anything in HiRes, then just copy, don't stretch
	if( DoubleHeight || DoubleWidth)
	{	
		Render16to16( Src, Dst);
		return;
	}

	lpDst = Dst.Surface;
	lpDst += ((Dst.Height - 478) >> 1) * Dst.Pitch;
    lpDst += (Dst.Width - 512) >> 1;

	for( H = 0; H != (ExtendHeight ? 448 : 478); H += 2)
	{
		DoubleLine8( (unsigned __int8 *)lpDst, (unsigned __int8 *)lpSrc, Src.Width);
		lpSrc += Src.Pitch;
		lpDst += Dst.Pitch;

		memset( lpDst, 0, 512);
		lpDst += Dst.Pitch;
	}

	if( ExtendHeight)
	for(; H != 478; H += 1)
	{	
		memset( lpDst, 0, 512);
		lpDst += Dst.Pitch;
	}
}


// Needed by all 8to16 functions
extern __int16 S9xPaletteWord[256];

// No enlargement, this just converts the whole 8-Bit mess to 16-bit
void Render8to16( SSurface Src, SSurface Dst)
{
    unsigned char *lpSrc, *lpDst;
    bool ExtendHeight, DoubleHeight, DoubleWidth;

    lpSrc = Src.Surface;

    ExtendHeight = Src.Height == 224 || Src.Height == 448;
    DoubleHeight = Src.Height > 240;
    DoubleWidth  = Src.Width == 512;

    lpDst = Dst.Surface;
    lpDst += ((Dst.Height - (DoubleHeight ? 478 : 239)) >> 1) * Dst.Pitch;
    lpDst += Dst.Width - Src.Width;
	
    for( unsigned int H = 0; H != Src.Height; H ++, lpDst += (Dst.Pitch-(Src.Width<<1)), lpSrc += (Src.Pitch-Src.Width))
    {	
        for( unsigned int W = 0; W != Src.Width; W++)
        {
            *((__int16 *)lpDst) = S9xPaletteWord[(*lpSrc)];

            lpDst += 2;
            lpSrc ++;
        }
    }

    if( ExtendHeight)
    {
        unsigned int HH = DoubleHeight ? 478 : 239;
        for(; H < HH; H++, lpDst += Dst.Pitch)
	{	memset( lpDst, 0, Src.Width<<1);	}
    }
}

// Enlarge 2x when the Snes9x is not rendering in HiRes
void RenderDouble8to16( SSurface Src, SSurface Dst)
{
unsigned char *lpSrc, *lpDst;
bool ExtendHeight, DoubleHeight, DoubleWidth;
unsigned int H;

	lpSrc = Src.Surface;

	ExtendHeight = Src.Height == 224 || Src.Height == 448;
	DoubleHeight = Src.Height > 240;
	DoubleWidth  = Src.Width == 512;

	// If Snes9x is rendering anything in HiRes, then just copy, don't stretch
	if( DoubleHeight || DoubleWidth)
	{	
		Render16to16( Src, Dst);
		return;
	}

	lpDst = Dst.Surface;
	lpDst += ((Dst.Height - 478) >> 1) * Dst.Pitch;
    lpDst += Dst.Width - 512;
	
	for( H = 0; H != Src.Height; H ++, lpDst += Dst.Pitch+(Dst.Pitch-(512<<1)), lpSrc += (Src.Pitch-256))
	{	
		for( unsigned int W = 0; W != Src.Width; W++)
		{
			*((__int16 *)(lpDst))             = S9xPaletteWord[(*lpSrc)];
			*((__int16 *)(lpDst+2))           = S9xPaletteWord[(*lpSrc)];
			*((__int16 *)(lpDst+Dst.Pitch))   = S9xPaletteWord[(*lpSrc)];
			*((__int16 *)(lpDst+2+Dst.Pitch)) = S9xPaletteWord[(*lpSrc)];

			lpDst += 4;
			lpSrc ++;
		}
	}

	if( ExtendHeight)
	for( H = 448; H != 478; H++, lpDst += Dst.Pitch)
	{	memset( lpDst, 0, 512<<1);	}
}

// Enlarge 2x using Fake TV mode when the Snes9x is not rendering in HiRes
// FakeTV mode = a black line between each other line
void RenderFakeTV8to16( SSurface Src, SSurface Dst)
{
unsigned char *lpSrc, *lpDst;
bool ExtendHeight, DoubleHeight, DoubleWidth;
unsigned int H;

	lpSrc = Src.Surface;

	ExtendHeight = Src.Height == 224 || Src.Height == 448;
	DoubleHeight = Src.Height > 240;
	DoubleWidth  = Src.Width == 512;

	// If Snes9x is rendering anything in HiRes, then just copy, don't stretch
	if( DoubleHeight || DoubleWidth)
	{	
		Render16to16( Src, Dst);
		return;
	}

	lpDst = Dst.Surface;
	lpDst += ((Dst.Height - 478) >> 1) * Dst.Pitch;
    lpDst += Dst.Width - 512;
	
	for( H = 0; H != Src.Height; H ++, lpDst += Dst.Pitch+(Dst.Pitch-(512<<1)), lpSrc += (Src.Pitch-256))
	{	
		memset( lpDst+Dst.Pitch, 0, 512<<1);
		for( unsigned int W = 0; W != Src.Width; W++)
		{
			*((__int16 *)lpDst)   = S9xPaletteWord[(*lpSrc)];
			*((__int16 *)(lpDst+2)) = S9xPaletteWord[(*lpSrc)];

			lpDst += 4;
			lpSrc ++;
		}
	}

	if( ExtendHeight)
	for( H = 448; H != 478; H++, lpDst += Dst.Pitch)
	{	memset( lpDst, 0, 512<<1);	}
}


// Needed by Render2xSaI16to16
//void SuperEagle( uint8 *srcPtr, uint32 srcPitch, uint8 *dstPtr, uint32 dstPitch, uint32 height, uint32 width);
void SuperEagle(uint8 *srcPtr, uint32 srcPitch, uint8 *deltaPtr, uint8 *dstPtr, uint32 dstPitch, int width, int height);
void _2xSaI(uint8 *srcPtr, uint32 srcPitch, uint8 *deltaPtr, uint8 *dstPtr, uint32 dstPitch, int width, int height);
void Super2xSaI(uint8 *srcPtr, uint32 srcPitch, uint8 *deltaPtr, uint8 *dstPtr, uint32 dstPitch, int width, int height);
bool mmx_checked = false;
bool mmx_enabled = true;
extern bool mmx_cpu;

// Needed by RenderDouble16to16, stretches a single line
inline void DoubleLine16( unsigned __int16 *lpDst, unsigned __int16 *lpSrc, unsigned int Width)
{
	for( unsigned int W = 0; W != Width; W ++)
	{
		lpDst[0] = lpSrc[0];
		lpDst[1] = lpSrc[0];

		lpDst += 2;
		lpSrc ++;
	}
}

// No enlargement, just render to the screen
void Render16to16( SSurface Src, SSurface Dst)
{
    unsigned char *lpSrc, *lpDst;
    bool ExtendHeight, DoubleHeight, DoubleWidth;

    lpSrc = Src.Surface;

    ExtendHeight = Src.Height == 224 || Src.Height == 448;
    DoubleHeight = Src.Height > 240;
    DoubleWidth  = Src.Width == 512;

    lpDst = Dst.Surface;
    lpDst += ((Dst.Height - (DoubleHeight ? 478 : 239)) >> 1) * Dst.Pitch;
    lpDst += Dst.Width - Src.Width;
	
    for( unsigned int H = 0; H != Src.Height; H ++, lpDst += Dst.Pitch, lpSrc += Src.Pitch)
    {	memcpy( lpDst, lpSrc, Src.Width<<1);	}

    if( ExtendHeight)
    {
        unsigned int HH = DoubleHeight ? 478 : 239;
	for(; H < HH; H++, lpDst += Dst.Pitch)
	{	memset( lpDst, 0, Src.Width<<1);	}
    }
}

// Enlarge 2x when the Snes9x is not rendering in HiRes
void RenderDouble16to16( SSurface Src, SSurface Dst)
{
    unsigned char *lpSrc, *lpDst;
    bool ExtendHeight, DoubleHeight, DoubleWidth;
    unsigned int H;

    lpSrc = Src.Surface;

    ExtendHeight = Src.Height == 224 || Src.Height == 448;
    DoubleHeight = Src.Height > 240;
    DoubleWidth  = Src.Width == 512;

    // If Snes9x is rendering anything in HiRes, then just copy, don't stretch
    if( DoubleHeight || DoubleWidth)
    {	
        Render16to16( Src, Dst);
        return;
    }

    lpDst = Dst.Surface;
    lpDst += ((Dst.Height - 478) >> 1) * Dst.Pitch;
    lpDst += Dst.Width - 512;

    unsigned int HH = ExtendHeight ? 448 : 478;
    for( H = 0; H < HH; H += 2)
    {
        DoubleLine16( (unsigned __int16 *)lpDst, (unsigned __int16 *)lpSrc, Src.Width);
        lpDst += Dst.Pitch;

        DoubleLine16( (unsigned __int16 *)lpDst, (unsigned __int16 *)lpSrc, Src.Width);
        lpSrc += Src.Pitch;
        lpDst += Dst.Pitch;
    }

    if( ExtendHeight)
	for(; H != 478; H += 1)
	{	
            memset( lpDst, 0, 512<<1);
            lpDst += Dst.Pitch;
	}
}

// Enlarge 2x using Fake TV mode when the Snes9x is not rendering in HiRes
// FakeTV mode = a black line between each other line
void RenderFakeTV16to16( SSurface Src, SSurface Dst)
{
    unsigned char *lpSrc, *lpDst;
    bool ExtendHeight, DoubleHeight, DoubleWidth;
    unsigned int H;

    lpSrc = Src.Surface;

    ExtendHeight = Src.Height == 224 || Src.Height == 448;
    DoubleHeight = Src.Height > 240;
    DoubleWidth  = Src.Width == 512;

    // If Snes9x is rendering anything in HiRes, then just copy, don't stretch
    if( DoubleHeight || DoubleWidth)
    {	
        Render16to16( Src, Dst);
        return;
    }

    lpDst = Dst.Surface;
    lpDst += ((Dst.Height - 478) >> 1) * Dst.Pitch;
    lpDst += Dst.Width - 512;

    unsigned int HH = ExtendHeight ? 448 : 478;
    for( H = 0; H < HH; H += 2)
    {
        DoubleLine16( (unsigned __int16 *)lpDst, (unsigned __int16 *)lpSrc, Src.Width);
        lpSrc += Src.Pitch;
        lpDst += Dst.Pitch;

        memset( lpDst, 0, 512*2);
        lpDst += Dst.Pitch;
    }

    if( ExtendHeight)
	for(; H != 478; H += 1)
	{	
            memset( lpDst, 0, 512<<1);
            lpDst += Dst.Pitch;
	}
}

// As the name suggests, I took this procedure from the MAME32, I hope that
// Michael Soderstrom and Chris Kirmse don't mind this.
static bool MAME32_DetectMMX()
{
#if defined(_M_IX86)

    bool result = false;
    int cpu_properties;


#define ASM_CPUID __asm _emit 0xf __asm _emit 0xa2

    __try
    {
        __asm
        {
            mov eax, 1
            ASM_CPUID
            mov cpu_properties, edx
        }
        /* if bit 23 is set, MMX is available */
        if (cpu_properties & 0x800000)
            result = true;
    }
    __except(1)
    {
        return false;
    }

    return result;

#else
    return false;
#endif
}

// Used as change log
unsigned char _2xSaI_Buffer[512*478*2];
// Needed when double buffering or when switching from another mode
unsigned char _2xSaI_TripleBuffer[512*478*2];
void RenderSuperEagle16to16( SSurface Src, SSurface Dst)
{
unsigned char *lpSrc, *lpDst;
bool ExtendHeight, DoubleHeight, DoubleWidth;

	if( !mmx_checked)
	{
		mmx_checked = true;
		mmx_cpu = false;

		if( mmx_enabled)
		  mmx_cpu = MAME32_DetectMMX();
	}

	lpSrc = Src.Surface;

	ExtendHeight = Src.Height == 224 || Src.Height == 448;
	DoubleHeight = Src.Height > 240;
	DoubleWidth  = Src.Width == 512;

	// If Snes9x is rendering anything in HiRes, then just copy, don't interpolate
	if( DoubleHeight || DoubleWidth)
	{	
		Render16to16( Src, Dst);
		return;
	}

	lpDst = Dst.Surface;
	lpDst += ((Dst.Height - 478) >> 1) * Dst.Pitch;
    lpDst += Dst.Width - 512;

	// Make sure the whole screen is redrawn when a rendermethod is switched
	if( OldRenderMethod != RenderMethod)
	{
		ZeroMemory( _2xSaI_Buffer, 512*478*2);
		ZeroMemory( _2xSaI_TripleBuffer, 512*478*2);
	}

	if( GUI.DoubleBuffered || OldRenderMethod != RenderMethod)
	{
		SuperEagle( lpSrc, Src.Pitch, _2xSaI_Buffer, _2xSaI_TripleBuffer, 512*2, Src.Width, Src.Height);

		unsigned char *S = _2xSaI_TripleBuffer;
		unsigned char *D = lpDst;
		for( int H = 0; H != (Src.Height << 1); H ++)
		{	
			memcpy( D, S, 512*2);
			D += Dst.Pitch;
			S += 512*2;
		}
	} 
	else
	{	SuperEagle( lpSrc, Src.Pitch, _2xSaI_Buffer, lpDst, Dst.Pitch, Src.Width, Src.Height);	}

	if( ExtendHeight)
	{
		lpDst += Dst.Pitch * 449;
		for(int H = 449; H != 478; H += 1)
		{	
			memset( lpDst, 0, 512<<1);
			lpDst += Dst.Pitch;
		}
	}
}

void Render2xSaI16to16( SSurface Src, SSurface Dst)
{
unsigned char *lpSrc, *lpDst;
bool ExtendHeight, DoubleHeight, DoubleWidth;

	if( !mmx_checked)
	{
		mmx_checked = true;
		mmx_cpu = false;

		if( mmx_enabled)
		  mmx_cpu = MAME32_DetectMMX();
	}

	lpSrc = Src.Surface;

	ExtendHeight = Src.Height == 224 || Src.Height == 448;
	DoubleHeight = Src.Height > 240;
	DoubleWidth  = Src.Width == 512;

	// If Snes9x is rendering anything in HiRes, then just copy, don't interpolate
	if( DoubleHeight || DoubleWidth)
	{	
		Render16to16( Src, Dst);
		return;
	}

	lpDst = Dst.Surface;
	lpDst += ((Dst.Height - 478) >> 1) * Dst.Pitch;
    lpDst += Dst.Width - 512;

	// Make sure the whole screen is redrawn when a rendermethod is switched

	if( OldRenderMethod != RenderMethod)
	{
		ZeroMemory( _2xSaI_Buffer, 512*478*2);
		ZeroMemory( _2xSaI_TripleBuffer, 512*478*2);
	}

	if( GUI.DoubleBuffered || OldRenderMethod != RenderMethod)
	{
		_2xSaI( lpSrc, Src.Pitch, _2xSaI_Buffer, _2xSaI_TripleBuffer, 512*2, Src.Width, Src.Height);

		unsigned char *S = _2xSaI_TripleBuffer;
		unsigned char *D = lpDst;
		for( int H = 0; H != (Src.Height << 1); H ++)
		{	
			memcpy( D, S, 512*2);
			D += Dst.Pitch;
			S += 512*2;
		}
	} 
	else
	{	_2xSaI( lpSrc, Src.Pitch, _2xSaI_Buffer, lpDst, Dst.Pitch, Src.Width, Src.Height);	}

	if( ExtendHeight)
	{
		lpDst += Dst.Pitch * 449;
		for(int H = 449; H != 478; H += 1)
		{	
			memset( lpDst, 0, 512<<1);
			lpDst += Dst.Pitch;
		}
	}
}

void RenderSuper2xSaI16to16( SSurface Src, SSurface Dst)
{
unsigned char *lpSrc, *lpDst;
bool ExtendHeight, DoubleHeight, DoubleWidth;

	if( !mmx_checked)
	{
		mmx_checked = true;
		mmx_cpu = false;

		if( mmx_enabled)
		  mmx_cpu = MAME32_DetectMMX();
	}

	lpSrc = Src.Surface;

	ExtendHeight = Src.Height == 224 || Src.Height == 448;
	DoubleHeight = Src.Height > 240;
	DoubleWidth  = Src.Width == 512;

	// If Snes9x is rendering anything in HiRes, then just copy, don't interpolate
	if( DoubleHeight || DoubleWidth)
	{	
		Render16to16( Src, Dst);
		return;
	}

	lpDst = Dst.Surface;
	lpDst += ((Dst.Height - 478) >> 1) * Dst.Pitch;
    lpDst += Dst.Width - 512;

	// Make sure the whole screen is redrawn when a rendermethod is switched

// The triple buffer scheme is disabled for Super2xSaI because it doesn't use the changes buffer yet
/*	if( OldRenderMethod != RenderMethod)
	{
		ZeroMemory( _2xSaI_Buffer, 512*478*2);
		ZeroMemory( _2xSaI_TripleBuffer, 512*478*2);
	}

	if( GUI.DoubleBuffered || OldRenderMethod != RenderMethod)
	{
		Super2xSaI( lpSrc, Src.Pitch, _2xSaI_Buffer, _2xSaI_TripleBuffer, 512*2, Src.Width, Src.Height);

		unsigned char *S = _2xSaI_TripleBuffer;
		unsigned char *D = lpDst;
		for( int H = 0; H != (Src.Height << 1); H ++)
		{	
			memcpy( D, S, 512*2);
			D += Dst.Pitch;
			S += 512*2;
		}
	} 
	else*/
	{	Super2xSaI( lpSrc, Src.Pitch, _2xSaI_Buffer, lpDst, Dst.Pitch, Src.Width, Src.Height);	}

	if( ExtendHeight)
	{
		lpDst += Dst.Pitch * 449;
		for(int H = 449; H != 478; H += 1)
		{	
			memset( lpDst, 0, 512<<1);
			lpDst += Dst.Pitch;
		}
	}
}
